import itertools
import logging
import os
import re
import shutil
import subprocess
import tempfile
from copy import deepcopy
from io import BytesIO

from docx.document import Document
from docx.opc.constants import CONTENT_TYPE
from docx.package import Package
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from lxml.etree import tostring

from bureaucracy.replacements import (HTMLReplacement, ImageReplacement,
                                      Replacement, TableReplacement,
                                      TextReplacement)
from bureaucracy.utils import namespaced

r = re.compile(r' MERGEFIELD +"?([^ ]+?)"? +(|\\\* MERGEFORMAT )', re.I)  # fixme. it might be not a simple as that
# the ooxml standard says that the instrText can be divided over several runs. i have not been able to make microsoft
# word generate something like this though...

logger = logging.getLogger('bureaucracy')


class DocxTemplate(Document):
    def __init__(self, docx, strict=False):
        """
        Initialize the DocxTemplate with the given docx file or stream.
        :param docx: A string or file like object (django.core.File objects work) representing a docx file.
        :param strict: will make the template render throw an error when fields and context do not match if True
        """
        self.strict = strict
        document_part = Package.open(docx).main_document_part
        if document_part.content_type != CONTENT_TYPE.WML_DOCUMENT_MAIN:
            tmpl = "file '%s' is not a Word file, content type is '%s'"
            raise ValueError(tmpl % (docx, document_part.content_type))
        super().__init__(document_part._element, document_part)

    def get_field_names(self):
        """
        Get the name of the mailmerge fields included in the document
        :return: a set of the fields' names
        """
        return {rv[0] for rv in self.iter_fields()}

    def iter_fields(self):
        """
        Gernerator for fldSimple and instrText fields and their fieldnames
        :return: a generator yielding tuples (field name, field)-tuples.
        """
        # there are two ways a mergefield can be represented: the simple way with fldSimple and the more
        # complex way with instrText and fldChar.

        simple_fields = ((field.attrib[namespaced('instr')], field) for field in self._element.xpath('.//w:fldSimple'))
        complex_fields = ((field.text, field) for field in self._element.xpath('.//w:instrText'))

        for instr, field in itertools.chain(simple_fields, complex_fields):

            m = r.match(instr)
            if not m and self.strict:
                raise ValueError("Could not determine name of merge field with instr text '{}'".format(instr))
            elif not m:
                logger.warning("Could not determine name of merge field with instr text '{}'. Skipping".format(instr))
                continue

            yield m.group(1), field

    def replace_simple_field(self, field, replacement):

        # a fldSimple tag is easily replaced, we just create a new run in the same paragraph and replace that one
        # with the fldSimple node.

        parent_node = field.getparent()

        # the standard says that this is the case most of the time so we only deal with this case for now:
        assert parent_node.tag == namespaced('p')

        current_paragraph = Paragraph(parent_node, self._body)
        replacement_run = Run(current_paragraph._p._add_r(), current_paragraph)
        parent_node.replace(field, replacement_run._element)
        replacement.fill(replacement_run)

    def replace_complex_field(self, field, replacement):
        # fldChar is more complex. it's not a tag, but rather a series of fldChar and instrText tags inside separate
        # runs. The tags that concern us are these:
        #
        #  1. <w:fldChar w:fldCharType="begin"/>: The beginning of the field
        #  2. <w:instrText xml:space="preserve"> MERGEFIELD test \* MERGEFORMAT </w:instrText> contains the field's name
        #  3. <w:fldChar w:fldCharType="end"/> Marks the end of the field
        #

        # get the run this field is in
        instr_run_node = field.getparent()
        assert instr_run_node.tag == namespaced('r')

        # we now look for the run containing of the opening fldChar for this instrText, which is the first one
        # with an opening fldChar we encounter before the run with instrText
        opening_run_node = instr_run_node
        while not opening_run_node.xpath('w:fldChar[@w:fldCharType="begin"]'):
            opening_run_node = opening_run_node.getprevious()
            if opening_run_node is None:
                raise ValueError(
                    "Could not find beginning of field with instr node '{}'?! Is the document malformed?".format(field))

        # idem for the run containing the closing fldChar, but of course now looking ahead
        closing_run_node = instr_run_node
        while not closing_run_node.xpath('w:fldChar[@w:fldCharType="end"]'):
            closing_run_node = closing_run_node.getnext()
            if closing_run_node is None:
                raise ValueError(
                    "Could not find end of field with instr node '{}'?! Is the document malformed?".format(field))

        # now replace all runs between the opening and closing runs
        current_paragraph = Paragraph(instr_run_node.getparent(), self._body)

        begin_index = current_paragraph._element.index(opening_run_node)
        end_index = current_paragraph._element.index(closing_run_node)

        run = Run(current_paragraph._p._add_r(), current_paragraph)
        current_paragraph._element[begin_index:end_index] = [run._element]

        replacement.fill(run)

    def replace_fields(self, context):
        unused_fields = set()
        unused_values = set(context.keys())

        for field_name, field in self.iter_fields():

            if field_name in context:
                unused_values.discard(field_name)
                replacement = context[field_name]
                if not isinstance(replacement, Replacement):
                    replacement = TextReplacement(replacement)
            else:
                if self.strict:
                    raise ValueError('Could not find field name {} in context'.format(field_name))
                else:
                    replacement = TextReplacement('')
                    unused_fields.add(field_name)

            if field.tag == namespaced('fldSimple'):
                self.replace_simple_field(field, replacement)
            elif field.tag == namespaced('instrText'):
                self.replace_complex_field(field, replacement)

        if unused_fields:
            logger.warn("Fields %s were present in the document, but not in the context. They were removed",
                        unused_fields)
        if unused_values:
            logger.warn(
                "Values %s were present in the context, but no corresponding fields were found in the document.",
                unused_values)

    def render(self, context, format='docx'):
        doc = deepcopy(self)  # take a copy so we can keep using this instance to generate from other contexts
        doc.replace_fields(context)

        if format == 'docx':
            handle = BytesIO()
            doc.save(handle)
            handle.seek(0)
            return handle.read()
        elif format == 'pdf':
            return doc.to_pdf_bytes()
        else:
            raise Exception('Unsupported format.')

    def render_and_save(self, path, context, format='docx'):
        doc = deepcopy(self)  # take a copy so we can keep using this instance to generate from other contexts
        doc.replace_fields(context)

        if format == 'docx':
            doc.save(path)
        elif format == 'pdf':
            doc.to_pdf(path)

    # what follows is a hack.
    # To be able to generate a pdf document/bytes from a merged document, we call libreoffice's headless
    # command line utility to to the generation. The problem is is that that tool only works with files,
    # not with bytestreams or strings, so we save this docx in a tmp dir, throw that into soffice, which
    # generates a pdf file in the same dir with the same name. Then we return bytes dumped into that file or move
    # the file to the desired path, depending on whether we're saving or rendering directly to a stream.
    # All this also makes rendering to pdf very slow. when doing a lot of renders, it might be a good idea to use
    # the batch functionality in the soffice convert tool.

    def _to_pdf(self, path=None):
        _, tmp_doc_path = tempfile.mkstemp(suffix='.docx')
        soffice_outdir = tempfile.gettempdir()
        tmp_pdf_outfilename = '{}.pdf'.format(tmp_doc_path.split(os.sep)[-1].split('.')[0])
        tmp_pdf_path = os.path.join(soffice_outdir, tmp_pdf_outfilename)

        self.save(tmp_doc_path)
        try:
            subprocess.call(['soffice', '--headless',
                             '--convert-to', 'pdf',
                             tmp_doc_path, '--outdir', soffice_outdir],
                            stdout=subprocess.DEVNULL)

            if path:
                shutil.move(tmp_pdf_path, path)
            else:
                with open(tmp_pdf_path, 'rb') as f:
                    return f.read()

        finally:
            if os.path.exists(tmp_doc_path):
                os.remove(tmp_doc_path)

            if os.path.exists(tmp_pdf_path):
                os.remove(tmp_pdf_path)

    def to_pdf(self, path):
        self._to_pdf(path)

    def to_pdf_bytes(self):
        return self._to_pdf()

    def __str__(self):
        return tostring(self._element, pretty_print=True).decode('utf-8')


if __name__ == '__main__':
    examples_dir = os.path.join(os.path.realpath(os.path.dirname(__file__)), '..', 'examples')
    doc = DocxTemplate(os.path.join(examples_dir, 'sample.docx'))
    context = {
        'table': TableReplacement([[str(i * j) for j in range(1, 8)] for i in range(1, 6)],
                                  headers=['one', 'two', 'three', 'four', 'five', 'six', 'seven']),
        'image': ImageReplacement(os.path.join(examples_dir, 'pigeony.jpg')),
        'html': HTMLReplacement(
            open(os.path.join(examples_dir, '..', 'examples', 'test.html')).read()),
        'text': 'some text',
        'hop': 546
    }

    doc.render_and_save(
        os.path.join(examples_dir, 'generated.docx'),
        context)



"""
This package implements the API to feed a powerpoint presentation as a template
to a source of data. The actual template contents can be templated out with
your favourite template engine.
"""

from .core import *  # noqa



class ShapeContainer:
    def __init__(self, shape):
        self.shape = shape
        self.children = []
        self.parents = []

    def wraps(self, other_shape):
        # other x coordinates must be less or equal than this ones
        if other_shape.x1 < self.x1 or other_shape.x2 > self.x2:
            return False
        # other y coordinates must be less or equal than this ones
        if other_shape.y1 < self.y1 or other_shape.y2 > self.y2:
            return False
        return True

    def add_child(self, other_shape):
        self.children.append(other_shape)
        other_shape.parents.append(self)

    @property
    def x1(self):
        return self.shape.left

    @property
    def x2(self):
        return self.shape.left + self.shape.width

    @property
    def y1(self):
        return self.shape.top

    @property
    def y2(self):
        return self.shape.top + self.shape.height

    @property
    def center_x(self):
        return self.shape.left + self.shape.width / 2

    @property
    def center_y(self):
        return self.shape.top + self.shape.height / 2

    @property
    def is_root(self):
        return not self.parents

    def get_placeholders(self):
        """
        Flatten the nested structure and return the placeholders in the correct order.
        """
        placeholders = [self.shape] if self.shape.is_placeholder else []
        children = sorted(self.children, key=lambda s: (s.center_y, s.center_x))
        for child in children:
            placeholders += child.get_placeholders()
        return placeholders



"""
This module defines the base engine to render template fragments.
"""


class BaseEngine:

    def render(self, fragment, context):
        raise NotImplementedError("You must implement the `render` method.")


class PythonEngine(BaseEngine):
    """
    Template engine that relies on python str.format(...).
    """

    def render(self, fragment, context):
        return fragment.format(**context)



from tests.powerpoint.test_templates import TEST_FILES

from bureaucracy.powerpoint import Template
from bureaucracy.powerpoint.placeholders import PlaceholderContainer
from bureaucracy.powerpoint.slides import SlideContainer


def test_insert_another_slide():
    test_file = str(TEST_FILES / 'template1.pptx')
    template = Template(test_file)

    assert len(template._presentation.slides) == 1

    slide = SlideContainer(template._presentation.slides[0], template._presentation)
    slide.insert_another()

    assert len(template._presentation.slides) == 2

    assert len(template._presentation.slides[0].shapes) == len(template._presentation.slides[1].shapes)

def test_insert_link():
    test_file = str(TEST_FILES / 'hyperlink.pptx')
    template = Template(test_file)

    ph = list(template._presentation.slides[0].placeholders)[0]
    shape = PlaceholderContainer(ph, "")

    shape.insert_link("http://www.whygodwhy.com", "foo")

    assert ph.text == "foo"
    assert ph.text_frame.paragraphs[0].runs[0].hyperlink.address == "http://www.whygodwhy.com"



import hashlib
from pathlib import Path
from unittest.mock import patch

from pptx import Presentation

from bureaucracy.powerpoint import Template
from bureaucracy.powerpoint.engines import BaseEngine, PythonEngine

TEST_FILES = Path(__file__).parent / 'files'


class ConstantEngine(BaseEngine):
    def render(self, fragment, context):
        return 'Constant'


class ContextObject:
    def __init__(self, link, desc):
        self.link = link
        self.desc = desc


def test_layouts_extraction():
    test_file = str(TEST_FILES / 'template1.pptx')
    template = Template(test_file)
    assert template.layouts == ['default', 'influencers']


def test_template_render(tmpdir):
    test_file = str(TEST_FILES / 'template1.pptx')
    template = Template(test_file)
    template.render(context={}, engine=ConstantEngine())

    outfile = str(tmpdir.join('constant-engine.pptx'))
    template.save_to(outfile)

    # check that the contents are correctly templated out
    pres = Presentation(outfile)
    assert len(pres.slides) == 1
    slide = pres.slides[0]
    assert slide.slide_layout.name == 'influencers'

    assert len(slide.placeholders) == 31 + 6
    for ph in slide.placeholders:
        assert ph.text == 'Constant'


def test_template_filled_in_placeholders(tmpdir):
    """
    Assert that filled in placeholders in the actual slide are not templated away.
    """
    test_file = str(TEST_FILES / 'empty-and-filled-in-placeholders.pptx')
    template = Template(test_file)
    context = {
        'language': 'Python',
    }
    template.render(context, engine=PythonEngine())
    outfile = str(tmpdir.join('placeholders.pptx'))
    template.save_to(outfile)

    # check that the contents are correctly templated out
    pres = Presentation(outfile)
    assert len(pres.slides) == 2

    # first slide has just plain template code in placeholders
    first_slide = pres.slides[0]
    assert len(first_slide.placeholders) == 4  # title, brand_logo, two placeholders

    ph1 = first_slide.slide_layout.placeholders[2].placeholder_format.idx
    ph2 = first_slide.slide_layout.placeholders[3].placeholder_format.idx

    # ids are fixed within a slide
    assert ph1 == 11
    assert ph2 == 12

    # compare rendered output
    assert first_slide.placeholders[ph1].text == 'A simple Python string format template'
    assert first_slide.placeholders[ph2].text == 'Another simple Python string format template'

    second_slide = pres.slides[1]
    assert first_slide.slide_layout == second_slide.slide_layout

    # compare rendered output
    assert second_slide.placeholders[ph1].text == 'Filled in placeholder – should not be replaced'
    assert second_slide.placeholders[ph2].text == 'Another simple Python string format template'


def test_control_placeholder(tmpdir):
    """
    Assert that control placeholders with zero-height are removed.

    A control placeholder can modify the context datastructure in place in
    prepration for the next placeholder(s). If it's deliberately set up to be
    zero height AND the output result is empty, it should be removed from the
    presentation alltogether to reduced clutter and confusion.
    """
    test_file = str(TEST_FILES / 'control-placeholder.pptx')
    template = Template(test_file)
    assert len(template._presentation.slides[0].placeholders) == 2

    context = {'control_placeholder_no_output': ''}
    template.render(context, engine=PythonEngine())
    outfile = str(tmpdir.join('control-ph.pptx'))
    template.save_to(outfile)

    # check that the contents are correctly templated out
    pres = Presentation(outfile)
    assert len(pres.slides) == 1

    slide = pres.slides[0]
    assert len(slide.placeholders) == 1

    ph_texts = [ph.text for ph in slide.placeholders]
    assert ph_texts == ['Click to edit Master title style']


def test_ph_ordering(tmpdir):
    """
    Assert that the placeholder template fragments are passed to the template
    engine in the right order.
    """
    test_file = str(TEST_FILES / 'ordering-placeholder.pptx')
    template = Template(test_file)
    assert len(template._presentation.slides) == 1
    context = {
        'first': 'First',
        'second': 'Second',
        'third': 'Third',
        'fourth': 'Fourth',
    }

    with patch.object(PythonEngine, 'render', return_value='some-string') as mocked_render:
        template.render(context, engine=PythonEngine())

    outfile = str(tmpdir.join('order-ph.pptx'))
    template.save_to(outfile)

    # check that the contents are correctly templated out
    pres = Presentation(outfile)
    assert len(pres.slides) == 1

    expected_calls = [
        (('{second}', context),),
        (('{third}', context),),
        (('{fourth}', context),),
        (('{first}', context),),
    ]
    assert mocked_render.call_args_list == expected_calls


def test_img_placeholder(tmpdir):
    test_file = str(TEST_FILES / 'simple_img.pptx')
    template = Template(test_file)
    assert len(template._presentation.slides[0].placeholders) == 1

    goat = str(TEST_FILES / 'goat.jpg')

    context = {
        'goat_here_pls': goat,
    }

    template.render(context, engine=PythonEngine())
    outfile = str(tmpdir.join('placeholders.pptx'))
    template.save_to(outfile)

    # check that the contents are correctly templated out

    pres = Presentation(outfile)
    assert len(pres.slides) == 1

    slide = pres.slides[0]
    # although the image file itself is placed elsewhere in the pptx zip, the image tag inside the slide tag
    # holds the image's sha1. checking that is enough for now.
    with open(goat, 'rb') as goat_file:
        expected_img_hexdigest = hashlib.sha1(goat_file.read()).hexdigest()
    assert expected_img_hexdigest == slide.shapes[0].image.sha1



from collections import OrderedDict

from pptx import Presentation
from pptx.slide import Slide

from .engines import BaseEngine
from .exceptions import TemplateSyntaxError
from .placeholders import PlaceholderContainer
from .shapes import ShapeContainer

CONTEXT_KEY_FOR_SLIDE = 'PPT_CURRENT_SLIDE'


class StopSlideRender(Exception):
    """
    Signals that a slide should stop rendering where it is now.
    """
    pass


class SlideContainer:
    def __init__(self, slide: Slide, presentation: Presentation):
        self.slide = slide
        self.presentation = presentation

    def __getattr__(self, name):
        """
        Proxy unknown attributes to the underlying slide object
        """
        return getattr(self.slide, name)

    def extract_shapes(self):
        """
        Extracts all the shapes from the slide and build a tree.

        Big shapes can wrap small shapes, and based on this we can re-group
        and nest them to apply an ordering to shapes (and thus placeholders).
        """
        if not self.slide.shapes:
            return []

        shapes = sorted(self.slide.slide_layout.shapes, key=lambda s: (s.width, s.height), reverse=True)
        wrapped_shapes = [ShapeContainer(shape) for shape in shapes]

        all_shapes = []

        while wrapped_shapes:
            shape = wrapped_shapes.pop(0)
            all_shapes.append(shape)
            for other_shape in wrapped_shapes:
                if not shape.wraps(other_shape):
                    continue
                shape.add_child(other_shape)

        # root shapes are the shapes without parent (= the biggest shapes that wrap other shapes)
        root_shapes = [shape for shape in all_shapes if shape.is_root]
        # finally, order by the center point of the shape
        root_shapes = sorted(root_shapes, key=lambda s: (s.center_y, s.center_x))
        return root_shapes

    def get_placeholder_idx_in_correct_order(self, fragments):
        """
        Given a slide, determine the order of template fragment evaluation.

        The slide placeholders are grouped by shapes which indicate that a
        set of placeholders needs to be evaluated before another set. This
        nesting translates into a deterministic order - top to bottom, and
        within a horizontal row from left to right.
        """
        shapes = self.extract_shapes()
        # we now have the correct order for the placeholders
        placeholders = sum((shape.get_placeholders() for shape in shapes), [])
        ordered_phs = [ph.placeholder_format.idx for ph in placeholders]
        return [idx for idx in ordered_phs if idx in fragments]

    def extract_template_code(self):
        """
        Extract the template code from slide placeholders.

        Placeholders have multiple 'levels': a placeholder can exist on a
        slide layout and on the slide itself. If the placeholder is filled in
        on the slide itself, it is not considered to be template code. If the
        value on the slide itself is empty, the value from the layout is taken
        and assumed to be template code.

        :return: an OrderedDict with placeholder id's as key and template code
          as value.
        """
        fragments = {}

        # set up the slide layout placeholder as template code
        for placeholder in self.slide.slide_layout.placeholders:
            fragments[placeholder.placeholder_format.idx] = placeholder.text

        # if a value exists for the placeholder in the slide itself, ignore the
        # template code
        for placeholder in self.slide.placeholders:
            if not placeholder.text:
                continue
            del fragments[placeholder.placeholder_format.idx]

        idxes = self.get_placeholder_idx_in_correct_order(fragments)
        # return the template bits in the right order
        return OrderedDict((idx, fragments[idx]) for idx in idxes)

    def render(self, engine: BaseEngine, context: dict):
        """
        Delegate the rendering to the underlying placeholders.
        """
        context[CONTEXT_KEY_FOR_SLIDE] = self

        fragments = self.extract_template_code()
        for idx, fragment in fragments.items():
            try:
                placeholder = PlaceholderContainer(self.slide.placeholders[idx], fragment)
            except KeyError:
                raise TemplateSyntaxError(
                    'Altough it is present on the slide master, the placeholder does '
                    'not seem to appear on the slide iself. Did you forget to apply the layout?'
                )
            try:
                placeholder.render(engine, context)
            except StopSlideRender:
                break
            finally:
                if placeholder.is_empty:
                    placeholder.remove()

    def insert_another(self):
        """
        Inserts the same slide into the presentation after the current position.

        NOTE: there's no insert slide method, only append to the end of the
        presentation, so we're using private API here.
        """
        layout = self.slide.slide_layout
        self.presentation.slides.add_slide(layout)

        current_index = self.presentation.slides.index(self.slide)
        # always at the end, and the ID is fixed (does not indicate position)
        # move the element to behind the current slide
        slide_list = self.presentation.slides._sldIdLst
        slide_list[current_index + 1] = slide_list[-1]
